import inspect
import logging

logger = logging.getLogger('pinspector')

class URIInvoker():

    def __init__(self, targets):
        self.targets = targets

    def convert(self, val:str, ref):
        obj = val
        if type(ref) is bool:
            obj = val.lower() in ['true', '1', 't', 'y', 'yes']
        elif type(ref) is int:
            obj = int(val)
        elif type(ref) is float:
            obj = float(val)
        return obj

    def getval(self, obj, key:str):
        try:
            if isinstance(obj, dict):
                return obj[key]
            elif isinstance(obj, (list, tuple)):
                return obj[int(key)]
        except:
            pass
        return getattr(obj, key, None)

    def setval(self, obj, key:str, val):
        if isinstance(obj, dict):
            obj[key] = val
        elif isinstance(obj, (list, tuple)):
            obj[int(key)] = val
        else:
            return setattr(obj, key, val)

    def isfield(self, obj):
        return not inspect.isroutine(obj)

    def invoke(self, target, spec:str, prev_target=None):
        if isinstance(target, str) and not prev_target:
            target = self.targets[target]
        if spec == '_length':
            return len(target)
        elif spec == '_type':
            return type(target).__name__
        elif spec == '_islist':
            return 'true' if isinstance(target, list) else 'false'
        elif spec == '_methods':
            return [f[0] + '() ' + type(f[1]).__name__ for f in inspect.getmembers(target, inspect.isroutine)]
        elif spec == '_fields':
            fields = [f[0] + ' ' + type(f[1]).__name__ for f in inspect.getmembers(target, self.isfield)]
            if hasattr(target, '__dict__'):
                fields += [k + ' ' + type(v).__name__ for k, v in vars(target).items()]
            return fields
        elif spec == '_keys':
            if isinstance(target, dict):
                return target.keys()
            else:
                return ''
        elif spec == '_protocols':
            protos = []
            if isinstance(target, dict):
                protos.append('dict')
            if isinstance(target, list):
                protos.append('list')
            return protos
        elif '..' in spec:
            spec_info = spec.split('..', 1)
            return self.invoke(self.invoke(target, spec_info[0], prev_target=prev_target), spec_info[1], prev_target=target)
        else:
            if spec.startswith('$'):
                spec = spec[1:]
            elif spec.startswith('~'):
                spec = spec[1:]
            if '~~' in spec:
                spec_info = spec.split('~~', 1)
                attr = self.getval(target, spec_info[0])
                if callable(attr):
                    args = spec_info[1].split('::')
                    try:
                        return attr(args)
                    except:
                        try:
                            return attr([int(arg) for arg in args])
                        except:
                            try:
                                return attr([eval(arg) for arg in args])
                            except:
                                return attr
                else:
                    val = self.convert(spec_info[1], attr)
                    self.setval(target, spec_info[0], val)
                    return self.getval(target, spec_info[0])
            else:
                attr = self.getval(target, spec)
                if isinstance(attr, (dict, list, int, str, float, bool)):
                    return attr
                elif callable(attr):
                    try:
                        return attr()
                    except:
                        return attr
                else:
                    return attr

    def print(self, obj):
        if isinstance(obj, str):
            return obj
        try:
            val = ''
            for el in obj:
                val += str(el)
                val += '\n'
            return val
        except:
            return str(obj)

    def run(self, target, spec:str):
        return self.print(self.invoke(target, spec))





from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer, HTTPServer
from http import HTTPStatus
import socketserver
from threading import Thread
from .invoker import URIInvoker
import logging
import copy
import inspect
import os

logger = logging.getLogger('pinspector')


class PinHandler(SimpleHTTPRequestHandler):

    def __init__(self, *args, pin_server=None, **kwargs):
        if pin_server is None:
            self.directory = None
        else:
            self.directory = pin_server.directory
        self.pin_server = pin_server
        super().__init__(*args, directory=self.directory, **kwargs)


    def do_GET(self):
        path = self.path.replace('/' + self.pin_server.prefix + '/', '/', 1)
        if path.startswith('/do/'):
            uri_info = path.split('/', 3)
            # from invoker
            res = self.pin_server.invoker.run(uri_info[2], uri_info[3])
            self.send_response_only(HTTPStatus.OK)
            self.end_headers()
            self.wfile.write(bytes(str(res), 'utf-8'))
        else:
            super().do_GET()


class PinServer():

    def __init__(self, port=5678, host='0.0.0.0', title='PIN - Python Inspector', targets=dict()):
        self.port = port
        self.host = host
        self.title = title
        self.prefix = 'pin'
        self.directory = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))) + '/static'
        self.started = False
        all_targets = {'service': self}
        if isinstance(targets, dict):
            for k, v in targets.items():
                all_targets[k] = v
        self.invoker = URIInvoker(all_targets)
        def finish_request(self, request, client_address):
            pin_server = getattr(self, 'pin_server', None)
            if pin_server:
                PinHandler(request, client_address, self, pin_server=pin_server)
            else:
                self.RequestHandlerClass(request, client_address, self)
        ThreadingHTTPServer.finish_request = finish_request

    def __call__(self):
        with ThreadingHTTPServer((self.host, self.port), PinHandler) as httpd:
            httpd.pin_server = self
            self.httpd = httpd
            httpd.serve_forever()

    def start(self):
        if not self.started:
            Thread(target=self).start()
            self.started = True

    def stop(self):
        if self.started:
            self.httpd.shutdown()
            self.started = False




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pinspector",
    version="0.8",
    author="Dinesh Ilindra",
    author_email="dinesh.iln@gmail.com",
    description="Live object inspector",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ilndinesh/pinspector",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data = True,
    python_requires='>=3.6',
)